"""@file func_layer.py
Main functions to encapsule oo calls
"""

# pylint: disable=too-many-lines
# pylint: disable=invalid-name
# Thin functonal layer on top of the class implementation of CLMModeling .
# The functions expect a global instance of the actual CLMModeling named
# "_modeling_object".

import numpy as np

if "_modeling_object" not in globals():
    _modeling_object = None


def compute_3d_density(
    r3d,
    mdelta,
    cdelta,
    z_cl,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    verbose=False,
    validate_input=True,
):
    r"""Retrieve the 3d density :math:`\rho(r)`.

    Profiles implemented so far are:

        `nfw`: :math:`\rho(r) = \frac{\rho_0}{\frac{c}{(r/R_{vir})}
        \left(1+\frac{c}{(r/R_{vir})}\right)^2}` (Navarro et al. 1996)

    Parameters
    ----------
    r3d : array_like, float
        Radial position from the cluster center in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster concentration
    z_cl: float
        Redshift of the cluster
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition; defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    verbose : boolean, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only available for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    rho : numpy.ndarray, float
        3-dimensional mass density in units of :math:`M_\odot\ Mpc^{-3}`

    Notes
    -----
    Need to refactor later so we only require arguments that are necessary for all profiles
    and use another structure to take the arguments necessary for specific models
    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    rho = _modeling_object.eval_3d_density(r3d, z_cl, verbose=verbose)

    _modeling_object.validate_input = True
    return rho


def compute_surface_density(
    r_proj,
    mdelta,
    cdelta,
    z_cl,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    r_mis=None,
    mis_from_backend=False,
    verbose=False,
    use_projected_quad=False,
    validate_input=True,
):
    r"""Computes the surface mass density

    .. math::
        \Sigma(R) = \int^\infty_{-\infty} dx\; \rho \left(\sqrt{R^2+x^2}\right),

    where :math:`\rho(r)` is the 3d density profile.

    If the `r_mis` keyword is specified, this function computes the miscentered surface
    density instead as

    .. math::
        \Sigma_{\rm mis}(R, R_{\rm mis}) = \frac{1}{2\pi}\int_0^{2\pi} d\theta \,
        \Sigma\left(\sqrt{R^2 + R_{\rm mis}^2 - 2 R R_{\rm mis} \cos\theta} \right)\;,


    Parameters
    ----------
    r_proj : array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster concentration
    z_cl: float
        Redshift of the cluster
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition; defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)

    r_mis : float, optional
        Projected miscenter distance in :math:`M\!pc`
    mis_from_backend : bool, optional
        If True, use the projected surface density from the backend for miscentering
        calculations. If False, use the (faster) CLMM exact analytical
        implementation instead. (Default: False)
    verbose : boolean, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only available for the NC and CCL
        backends.
    use_projected_quad : bool
        Only available for Einasto profile with CCL as the backend. If True, CCL will use
        quad_vec instead of default FFTLog to calculate the surface density profile.
        Default: False
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    sigma : numpy.ndarray, float
        2D projected surface density in units of :math:`M_\odot\ Mpc^{-2}`

    Notes
    -----
    Need to refactory so we only require arguments that are necessary for all models and use
    another structure to take the arguments necessary for specific models.
    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)
    if halo_profile_model == "einasto" and _modeling_object.backend == "ccl":
        _modeling_object.set_projected_quad(use_projected_quad)

    sigma = _modeling_object.eval_surface_density(
        r_proj, z_cl, r_mis=r_mis, mis_from_backend=mis_from_backend, verbose=verbose
    )

    _modeling_object.validate_input = True
    return sigma


def compute_mean_surface_density(
    r_proj,
    mdelta,
    cdelta,
    z_cl,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    r_mis=None,
    mis_from_backend=False,
    verbose=False,
    validate_input=True,
):
    r"""Computes the mean value of surface density inside radius `r_proj`

    .. math::
        \bar{\Sigma}(<R) = \frac{2}{R^2} \int^R_0 dR' R' \Sigma(R'),

    Parameters
    ----------
    r_proj : array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster concentration
    z_cl: float
        Redshift of the cluster
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition; defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope. Option only
        available for the NumCosmo backend
    r_mis : float, optional
        Projected miscenter distance in :math:`M\!pc`
    mis_from_backend : bool, optional
        If True, use the projected surface density from the backend for miscentering
        calculations. If False, use the (faster) CLMM exact analytical
        implementation instead. (Default: False)
    verbose : boolean, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only available for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.


    Returns
    -------
    sigma : numpy.ndarray, float
        2D projected surface density in units of :math:`M_\odot\ Mpc^{-2}`

    Notes
    -----
    Need to refactory so we only require arguments that are necessary for all models and use
    another structure to take the arguments necessary for specific models.
    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    sigma_bar = _modeling_object.eval_mean_surface_density(
        r_proj, z_cl, r_mis=r_mis, mis_from_backend=mis_from_backend, verbose=verbose
    )

    _modeling_object.validate_input = True
    return sigma_bar


def compute_excess_surface_density(
    r_proj,
    mdelta,
    cdelta,
    z_cl,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    r_mis=None,
    mis_from_backend=False,
    verbose=False,
    validate_input=True,
):
    r"""Computes the excess surface density

    .. math::
        \Delta\Sigma(R) = \bar{\Sigma}(<R)-\Sigma(R),

    Parameters
    ----------
    r_proj : array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster concentration
    z_cl: float
        Redshift of the cluster
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition; defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    r_mis : float, optional
        Projected miscenter distance in :math:`M\!pc`
    mis_from_backend : bool, optional
        If True, use the projected surface density from the backend for miscentering
        calculations. If False, use the (faster) CLMM exact analytical
        implementation instead. (Default: False)
    verbose : boolean, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only available for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    deltasigma : numpy.ndarray, float
        Excess surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    deltasigma = _modeling_object.eval_excess_surface_density(
        r_proj, z_cl, r_mis=r_mis, mis_from_backend=mis_from_backend, verbose=verbose
    )

    _modeling_object.validate_input = True
    return deltasigma


def compute_excess_surface_density_2h(
    r_proj,
    z_cl,
    cosmo,
    halobias=1.0,
    logkbounds=(-5, 5),
    ksteps=1000,
    loglbounds=(0, 6),
    lsteps=500,
    validate_input=True,
):
    r"""Computes the 2-halo term excess surface density from eq.(13) of Oguri & Hamana (2011)

    .. math::
        \Delta\Sigma_{\text{2h}}(R) = \frac{\rho_m(z)b(M)}{(1 + z)^3D_A(z)^2}
        \int\frac{ldl}{(2\pi)} P_{\text{mm}}(k_l, z)J_2(l\theta)

    where

    .. math::
        k_l = \frac{l}{D_A(z)(1 +z)}

    and :math:`b(M)` is the halo bias

    Parameters
    ----------
    r_proj : array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    z_cl: float
        Redshift of the cluster
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    halobias : float, optional
        Value of the halo bias
    logkbounds : tuple(float,float), shape(2,), optional
        Log10 of the upper and lower bounds for the linear matter power spectrum
    ksteps : int, optional
        Number of steps in k-space
    loglbounds : tuple(float,float), shape(2,), optional
        Log10 of the upper and lower bounds for numerical integration
    lsteps : int, optional
        Steps for the numerical integration
    validate_input: bool
        Validade each input argument

    Returns
    -------
    deltasigma_2h : numpy.ndarray, float
        2-halo term excess surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)

    deltasigma_2h = _modeling_object.eval_excess_surface_density_2h(
        r_proj,
        z_cl,
        halobias=halobias,
        logkbounds=logkbounds,
        ksteps=ksteps,
        loglbounds=loglbounds,
        lsteps=lsteps,
    )

    _modeling_object.validate_input = True
    return deltasigma_2h


def compute_surface_density_2h(
    r_proj,
    z_cl,
    cosmo,
    halobias=1,
    logkbounds=(-5, 5),
    ksteps=1000,
    loglbounds=(0, 6),
    lsteps=500,
    validate_input=True,
):
    r"""Computes the 2-halo term surface density from eq.(13) of Oguri & Hamana (2011)

    .. math::
        \Sigma_{\text{2h}}(R) = \frac{\rho_\text{m}(z)b(M)}{(1 + z)^3D_A(z)^2}
        \int\frac{ldl}{(2\pi)}P_{\text{mm}}(k_l, z)J_0(l\theta)

    where

    .. math::
        k_l = \frac{l}{D_A(z)(1 + z)}

    and :math:`b(M)` is the halo bias

    Parameters
    ----------
    r_proj : array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    z_cl: float
        Redshift of the cluster
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    halobias : float, optional
        Value of the halo bias
    logkbounds : tuple(float,float), shape(2,), optional
        Log10 of the upper and lower bounds for the linear matter power spectrum
    ksteps : int, optional
        Number of steps in k-space
    loglbounds : tuple(float,float), shape(2,), optional
        Log10 of the upper and lower bounds for numerical integration
    lsteps : int, optional
        Steps for the numerical integration
    validate_input: bool
        Validade each input argument

    Returns
    -------
    sigma_2h : numpy.ndarray, float
        2-halo term surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)

    sigma_2h = _modeling_object.eval_surface_density_2h(
        r_proj,
        z_cl,
        halobias=halobias,
        logkbounds=logkbounds,
        ksteps=ksteps,
        loglbounds=loglbounds,
        lsteps=lsteps,
    )

    _modeling_object.validate_input = True
    return sigma_2h


def compute_critical_surface_density_eff(cosmo, z_cluster, pzbins, pzpdf, validate_input=True):
    r"""Computes the 'effective critical surface density'

    .. math::
        \langle \Sigma_{\text{crit}}^{-1}\rangle^{-1} =
        \left(\int \frac{1}{\Sigma_{\text{crit}}(z)}p(z) \mathrm{d}z\right)^{-1}

    where :math:`p(z)` is the source photoz probability density function.
    This comes from the maximum likelihood estimator for evaluating a :math:`\Delta\Sigma`
    profile.

    For the standard :math:`\Sigma_{\text{crit}}(z)` definition, use the `eval_sigma_crit` method of
    the CLMM cosmology object.

    Parameters
    ----------
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    z_cluster : float
        Galaxy cluster redshift
    pzbins : array-like
        Bins where the source redshift pdf is defined
    pzpdf : array-like
        Values of the source redshift pdf
    validate_input: bool
        Validade each input argument


    Returns
    -------
    sigma_c : numpy.ndarray, float
        Cosmology-dependent (effective) critical surface density in units of
        :math:`M_\odot\ Mpc^{-2}`
    """

    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    sigma_c = _modeling_object.eval_critical_surface_density_eff(z_cluster, pzbins, pzpdf)

    _modeling_object.validate_input = True
    return sigma_c


def compute_tangential_shear(
    r_proj,
    mdelta,
    cdelta,
    z_cluster,
    z_src,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    z_src_info="discrete",
    verbose=False,
    validate_input=True,
):
    r"""Computes the tangential shear

    .. math::
        \gamma_t = \frac{\Delta\Sigma}{\Sigma_{crit}} = \frac{\bar{\Sigma}-\Sigma}{\Sigma_{crit}}

    or

    .. math::
        \gamma_t = \gamma_\infty \times \beta_s

    Parameters
    ----------
    r_proj : array_like, float
        The projected radial positions in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster NFW concentration.
    z_cluster : float
        Galaxy cluster redshift
    z_src : array_like, float, function
        Information on the background source galaxy redshift(s). Value required depends on
        `z_src_info` (see below).
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition.  Defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    z_src_info : str, optional
        Type of redshift information provided by the `z_src` argument.
        The following supported options are:

            * 'discrete' (default) : The redshift of sources is provided by `z_src`.
              It can be individual redshifts for each source galaxy when `z_src` is an array
              or all sources are at the same redshift when `z_src` is a float.

            * 'distribution' : A redshift distribution function is provided by `z_src`.
              `z_src` must be a one dimensional function.

            * 'beta' : The averaged lensing efficiency is provided by `z_src`.
              `z_src` must be a tuple containing
              ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
              the lensing efficiency and square of the lensing efficiency averaged over
              the galaxy redshift distribution repectively.

                .. math::
                    \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                    {D_{L,\infty}}\right\rangle

                .. math::
                    \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                    {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

    verbose : bool, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    gammat : numpy.ndarray, float
        Tangential shear
    """

    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)
    if np.min(r_proj) < 1.0e-11:
        raise ValueError(
            f"Rmin = {np.min(r_proj):.2e} Mpc/h! This value is too small "
            "and may cause computational issues."
        )

    tangential_shear = _modeling_object.eval_tangential_shear(
        r_proj,
        z_cluster,
        z_src,
        z_src_info=z_src_info,
        verbose=verbose,
    )

    _modeling_object.validate_input = True
    return tangential_shear


def compute_convergence(
    r_proj,
    mdelta,
    cdelta,
    z_cluster,
    z_src,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    z_src_info="discrete",
    verbose=False,
    validate_input=True,
):
    r"""Computes the mass convergence

    .. math::
        \kappa = \frac{\Sigma}{\Sigma_{crit}}

    or

    .. math::
        \kappa = \kappa_\infty \times \beta_s

    Parameters
    ----------
    r_proj : array_like, float
        The projected radial positions in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster NFW concentration.
    z_cluster : float
        Galaxy cluster redshift
    z_src : array_like, float, function
        Information on the background source galaxy redshift(s). Value required depends on
        `z_src_info` (see below).
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition.  Defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    z_src_info : str, optional
        Type of redshift information provided by the `z_src` argument.
        The following supported options are:

            * 'discrete' (default) : The redshift of sources is provided by `z_src`.
              It can be individual redshifts for each source galaxy when `z_src` is an array
              or all sources are at the same redshift when `z_src` is a float.

            * 'distribution' : A redshift distribution function is provided by `z_src`.
              `z_src` must be a one dimensional function.

            * 'beta' : The averaged lensing efficiency is provided by `z_src`.
              `z_src` must be a tuple containing
              ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
              the lensing efficiency and square of the lensing efficiency averaged over
              the galaxy redshift distribution repectively.

                .. math::
                    \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                    {D_{L,\infty}}\right\rangle

                .. math::
                    \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                    {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

    verbose : bool, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    kappa : numpy.ndarray, float
        Mass convergence, kappa.

    """

    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    convergence = _modeling_object.eval_convergence(
        r_proj,
        z_cluster,
        z_src,
        z_src_info=z_src_info,
        verbose=verbose,
    )

    _modeling_object.validate_input = True
    return convergence


def compute_reduced_tangential_shear(
    r_proj,
    mdelta,
    cdelta,
    z_cluster,
    z_src,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    z_src_info="discrete",
    approx=None,
    integ_kwargs=None,
    alpha_ein=None,
    validate_input=True,
    verbose=False,
):
    r"""Computes the reduced tangential shear

    .. math::
        g_t = \frac{\gamma_t}{1-\kappa}

    Parameters
    ----------
    r_proj : array_like, float
        The projected radial positions in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster NFW concentration.
    z_cluster : float
        Galaxy cluster redshift
    z_src : array_like, float, function
        Information on the background source galaxy redshift(s). Value required depends on
        `z_src_info` (see below).
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition.  Defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default);
            * 'critical' - not in cluster_toolkit;
            * 'virial' - not in cluster_toolkit;

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    z_src_info : str, optional
        Type of redshift information provided by the `z_src` argument.
        The following supported options are:

            * 'discrete' (default) : The redshift of sources is provided by `z_src`.
              It can be individual redshifts for each source galaxy when `z_src` is an array
              or all sources are at the same redshift when `z_src` is a float
              (Used for `approx=None`).

            * 'distribution' : A redshift distribution function is provided by `z_src`.
              `z_src` must be a one dimensional function (Used when `approx=None`).

            * 'beta' : The averaged lensing efficiency is provided by `z_src`.
              `z_src` must be a tuple containing
              ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
              the lensing efficiency and square of the lensing efficiency averaged over
              the galaxy redshift distribution repectively.

                .. math::
                    \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                    {D_{L,\infty}}\right\rangle

                .. math::
                    \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                    {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

    approx : str, optional
        Type of computation to be made for reduced tangential shears, options are:

            * None (default): Requires `z_src_info` to be 'discrete' or 'distribution'.
              If `z_src_info='discrete'`, full computation is made for each
              `r_proj, z_src` pair individually. If `z_src_info='distribution'`, reduced
              tangential shear at each value of `r_proj` is calculated as

              .. math::
                  g_t
                  =\left<\frac{\beta_s\gamma_{\infty}}{1-\beta_s\kappa_{\infty}}\right>
                  =\frac{\int_{z_{min}}^{z_{max}}\frac{\beta_s(z)\gamma_{\infty}}
                  {1-\beta_s(z)\kappa_{\infty}}N(z)\text{d}z}
                  {\int_{z_{min}}^{z_{max}} N(z)\text{d}z}

            * 'order1' : Same approach as in Weighing the Giants - III (equation 6 in
              Applegate et al. 2014; https://arxiv.org/abs/1208.0605).
              `z_src_info` must be 'beta':

              .. math::
                  g_t\approx\frac{\left<\beta_s\right>\gamma_{\infty}}
                  {1-\left<\beta_s\right>\kappa_{\infty}}

            * 'order2' : Same approach as in Cluster Mass Calibration at High
              Redshift (equation 12 in Schrabback et al. 2017;
              https://arxiv.org/abs/1611.03866).
              `z_src_info` must be 'beta':

              .. math::
                  g_t\approx\frac{\left<\beta_s\right>\gamma_{\infty}}
                  {1-\left<\beta_s\right>\kappa_{\infty}}
                  \left(1+\left(\frac{\left<\beta_s^2\right>}
                  {\left<\beta_s\right>^2}-1\right)\left<\beta_s\right>\kappa_{\infty}\right)

    integ_kwargs: None, dict
        Extra arguments for the redshift integration (when
        `approx=None, z_src_info='distribution'`). Possible keys are:

            * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
              when performing the sum. (default=None)
            * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
              when performing the sum. (default=10.0)
            * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
              `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    verbose : bool, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    gt : numpy.ndarray, float
        Reduced tangential shear

    """
    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    red_tangential_shear = _modeling_object.eval_reduced_tangential_shear(
        r_proj,
        z_cluster,
        z_src,
        z_src_info=z_src_info,
        approx=approx,
        integ_kwargs=integ_kwargs,
        verbose=verbose,
    )

    _modeling_object.validate_input = True
    return red_tangential_shear


def compute_magnification(
    r_proj,
    mdelta,
    cdelta,
    z_cluster,
    z_src,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    z_src_info="discrete",
    approx=None,
    integ_kwargs=None,
    verbose=False,
    validate_input=True,
):
    r"""Computes the magnification

    .. math::
        \mu = \frac{1}{(1-\kappa)^2-|\gamma_t|^2}

    Parameters
    ----------
    r_proj : array_like, float
        The projected radial positions in :math:`M\!pc`.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster NFW concentration.
    z_cluster : float
        Galaxy cluster redshift
    z_src : array_like, float, function
        Information on the background source galaxy redshift(s). Value required depends on
        `z_src_info` (see below).
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition.  Defaults to 200.
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    z_src_info : str, optional
        Type of redshift information provided by the `z_src` argument.
        The following supported options are:

            * 'discrete' (default) : The redshift of sources is provided by `z_src`.
              It can be individual redshifts for each source galaxy when `z_src` is an array
              or all sources are at the same redshift when `z_src` is a float
              (Used for `approx=None`).

            * 'distribution' : A redshift distribution function is provided by `z_src`.
              `z_src` must be a one dimensional function (Used when `approx=None`).

            * 'beta' : The averaged lensing efficiency is provided by `z_src`.
              `z_src` must be a tuple containing
              ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
              the lensing efficiency and square of the lensing efficiency averaged over
              the galaxy redshift distribution repectively.

                .. math::
                    \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                    {D_{L,\infty}}\right\rangle

                .. math::
                    \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                    {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

    approx : str, optional
        Type of computation to be made for magnifications, options are:

            * None (default): Requires `z_src_info` to be 'discrete' or 'distribution'.
              If `z_src_info='discrete'`, full computation is made for each
              `r_proj, z_src` pair individually. If `z_src_info='distribution'`, magnification
              at each value of `r_proj` is calculated as

              .. math::
                  \mu
                  =\left<\frac{1}{\left(1-\beta_s\kappa_{\infty}\right)^2
                  -\left(\beta_s\gamma_{\infty}\right)^2}\right>
                  =\frac{\int_{z_{min}}^{z_{max}}\frac{N(z)\text{d}z}
                  {\left(1-\beta_s(z)\kappa_{\infty}\right)^2
                  -\left(\beta_s(z)\gamma_{\infty}\right)^2}}
                  {\int_{z_{min}}^{z_{max}} N(z)\text{d}z}

            * 'order1' : Uses the weak lensing approximation of the magnification with up to
              first-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
              (`z_src_info` must be 'beta'):

              .. math::
                  \mu \approx 1 + 2 \left<\beta_s\right>\kappa_{\infty}

            * 'order2' : Uses the weak lensing approximation of the magnification with up to
              second-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
              (`z_src_info` must be 'beta'):

              .. math::
                  \mu \approx 1 + 2 \left<\beta_s\right>\kappa_{\infty}
                  + 3 \left<\beta_s^2\right>\kappa_{\infty}^2
                  + \left<\beta_s^2\right>\gamma_{\infty}^2

    integ_kwargs: None, dict
        Extra arguments for the redshift integration (when
        `approx=None, z_src_info='distribution'`). Possible keys are:

            * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
              when performing the sum. (default=None)
            * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
              when performing the sum. (default=10.0)
            * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
              `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

    verbose : bool, optional
        If True, the Einasto slope (alpha_ein) is printed out. Only availble for the NC and CCL
        backends.
    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    magnification : numpy.ndarray, float
        Magnification :math:`\mu`.

    """

    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    magnification = _modeling_object.eval_magnification(
        r_proj,
        z_cluster,
        z_src,
        z_src_info=z_src_info,
        approx=approx,
        integ_kwargs=integ_kwargs,
        verbose=verbose,
    )

    _modeling_object.validate_input = True
    return magnification


def compute_magnification_bias(
    r_proj,
    alpha,
    mdelta,
    cdelta,
    z_cluster,
    z_src,
    cosmo,
    delta_mdef=200,
    halo_profile_model="nfw",
    massdef="mean",
    alpha_ein=None,
    z_src_info="discrete",
    approx=None,
    integ_kwargs=None,
    verbose=False,
    validate_input=True,
):
    r""" Computes magnification bias from magnification :math:`\mu`
    and slope parameter :math:`\alpha` as :

    .. math::
        \mu^{\alpha - 1}.

    The alpha parameter depends on the source sample and is computed as the slope of the
    cummulative numer counts at a given magnitude :

    .. math::
        \alpha \equiv \alpha(f) = - \frac{\mathrm{d}}{\mathrm{d}\log{f}} \log{n_0(>f)}

    or,

    .. math::
        \alpha \equiv \alpha(m) = 2.5 \frac{\mathrm d}{\mathrm d m} \log{n_0(<m)}

    see e.g.  Bartelmann & Schneider 2001; Umetsu 2020

    Parameters
    ----------
    r_proj : array_like, float
        The projected radial positions in :math:`M\!pc`.
    alpha : array like
        The slope of the cummulative source number counts.
    mdelta : float
        Galaxy cluster mass in :math:`M_\odot`.
    cdelta : float
        Galaxy cluster NFW concentration.
    z_cluster : float
        Galaxy cluster redshift
    z_src : array_like, float, function
        Information on the background source galaxy redshift(s). Value required depends on
        `z_src_info` (see below).
    cosmo : clmm.cosmology.Cosmology object
        CLMM Cosmology object
    delta_mdef : int, optional
        Mass overdensity definition.  Defaults to 200.
    alpha_ein : float, None, optional
        If `halo_profile_model=='einasto'`, set the value of the Einasto slope.
        Option only available for the NumCosmo and CCL backends.
        If None, use the default value of the backend. (0.25 for the NumCosmo backend and a
        cosmology-dependent value for the CCL backend.)
    halo_profile_model : str, optional
        Profile model parameterization (letter case independent):

            * 'nfw' (default)
            * 'einasto' - not in cluster_toolkit
            * 'hernquist' - not in cluster_toolkit

    massdef : str, optional
        Profile mass definition, with the following supported options (letter case independent):

            * 'mean' (default)
            * 'critical'
            * 'virial'

    z_src_info : str, optional
        Type of redshift information provided by the `z_src` argument.
        The following supported options are:

            * 'discrete' (default) : The redshift of sources is provided by `z_src`.
              It can be individual redshifts for each source galaxy when `z_src` is an array
              or all sources are at the same redshift when `z_src` is a float
              (Used for `approx=None`).

            * 'distribution' : A redshift distribution function is provided by `z_src`.
              `z_src` must be a one dimensional function (Used when `approx=None`).

            * 'beta' : The averaged lensing efficiency is provided by `z_src`.
              `z_src` must be a tuple containing
              ( :math:`\langle \beta_s \rangle, \langle \beta_s^2 \rangle`),
              the lensing efficiency and square of the lensing efficiency averaged over
              the galaxy redshift distribution repectively.

                .. math::
                    \langle \beta_s \rangle = \left\langle \frac{D_{LS}}{D_S}\frac{D_\infty}
                    {D_{L,\infty}}\right\rangle

                .. math::
                    \langle \beta_s^2 \rangle = \left\langle \left(\frac{D_{LS}}
                    {D_S}\frac{D_\infty}{D_{L,\infty}}\right)^2 \right\rangle

    approx : str, optional
        Type of computation to be made for magnification biases, options are:

            * None (default): Requires `z_src_info` to be 'discrete' or 'distribution'.
              If `z_src_info='discrete'`, full computation is made for each
              `r_proj, z_src` pair individually. If `z_src_info='distribution'`, magnification
              bias at each value of `r_proj` is calculated as

              .. math::
                  \mu^{\alpha-1}
                  &=\left(\left<\frac{1}{\left(1-\beta_s\kappa_{\infty}\right)^2
                  -\left(\beta_s\gamma_{\infty}\right)^2}\right>\right)^{\alpha-1}
                  \\\\
                  &=\frac{\int_{z_{min}}^{z_{max}}\frac{N(z)\text{d}z}
                  {\left(\left(1-\beta_s(z)\kappa_{\infty}\right)^2
                  -\left(\beta_s(z)\gamma_{\infty}\right)^2\right)^{\alpha-1}}}
                  {\int_{z_{min}}^{z_{max}} N(z)\text{d}z}

            * 'order1' : Uses the weak lensing approximation of the magnification bias with up
              to first-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
              (`z_src_info` must be 'beta'):

              .. math::
                  \mu^{\alpha-1} \approx
                  1 + \left(\alpha-1\right)\left(2 \left<\beta_s\right>\kappa_{\infty}\right)

            * 'order2' : Uses the weak lensing approximation of the magnification bias with up
              to second-order terms in :math:`\kappa_{\infty}` or :math:`\gamma_{\infty}`
              `z_src_info` must be 'beta':

              .. math::
                  \mu^{\alpha-1} \approx
                  1 &+ \left(\alpha-1\right)\left(2 \left<\beta_s\right>\kappa_{\infty}\right)
                  \\\\
                  &+ \left(\alpha-1\right)\left(\left<\beta_s^2\right>\gamma_{\infty}^2\right)
                  \\\\
                  &+ \left(2\alpha-1\right)\left(\alpha-1\right)
                  \left(\left<\beta_s^2\right>\kappa_{\infty}^2\right)

    integ_kwargs: None, dict
        Extra arguments for the redshift integration (when
        `approx=None, z_src_info='distribution'`). Possible keys are:

            * 'zmin' (None, float) : Minimum redshift to be set as the source of the galaxy
              when performing the sum. (default=None)
            * 'zmax' (float) : Maximum redshift to be set as the source of the galaxy
              when performing the sum. (default=10.0)
            * 'delta_z_cut' (float) : Redshift cut so that `zmin` = `z_cl` + `delta_z_cut`.
              `delta_z_cut` is ignored if `z_min` is already provided. (default=0.1)

    validate_input : bool, optional
        If True (default), the types of the arguments are checked before proceeding.

    Returns
    -------
    magnification_bias : numpy.ndarray
        magnification bias
    """

    _modeling_object.validate_input = validate_input
    _modeling_object.set_cosmo(cosmo)
    _modeling_object.set_halo_density_profile(
        halo_profile_model=halo_profile_model, massdef=massdef, delta_mdef=delta_mdef
    )
    _modeling_object.set_concentration(cdelta)
    _modeling_object.set_mass(mdelta)
    if halo_profile_model == "einasto" or alpha_ein is not None:
        _modeling_object.set_einasto_alpha(alpha_ein)

    magnification_bias = _modeling_object.eval_magnification_bias(
        r_proj,
        z_cluster,
        z_src,
        alpha,
        z_src_info=z_src_info,
        approx=approx,
        integ_kwargs=integ_kwargs,
        verbose=verbose,
    )

    _modeling_object.validate_input = True
    return magnification_bias



"""CLMM is a cluster mass modeling code."""

from . import support
from .clusterensemble import ClusterEnsemble
from .dataops import compute_tangential_and_cross_components, make_radial_profile
from .galaxycluster import GalaxyCluster
from .gcdata import GCData
from .theory import (
    Cosmology,
    Modeling,
    compute_3d_density,
    compute_convergence,
    compute_excess_surface_density,
    compute_excess_surface_density_2h,
    compute_magnification,
    compute_magnification_bias,
    compute_magnification_bias_from_magnification,
    compute_mean_surface_density,
    compute_profile_mass_in_radius,
    compute_rdelta,
    compute_reduced_shear_from_convergence,
    compute_reduced_tangential_shear,
    compute_surface_density,
    compute_surface_density_2h,
    compute_tangential_shear,
    convert_profile_mass_concentration,
)
from .utils import compute_radial_averages, convert_units, make_bins

__version__ = "1.16.5"



"""General utility functions that are used in multiple modules"""

import numpy as np
from scipy.integrate import quad

from . import redshift_distributions as zdist


def compute_beta(z_src, z_cl, cosmo):
    r"""Geometric lensing efficicency

    .. math::
        \beta = max(0, D_{a,\ ls}/D_{a,\ s})

    Eq.2 in https://arxiv.org/pdf/1611.03866.pdf

    Parameters
    ----------
    z_src : float, array_like
        Source galaxy redshift
    z_cl: float
        Galaxy cluster redshift
    cosmo: clmm.Cosmology
        CLMM Cosmology object

    Returns
    -------
    float, array
        Geometric lensing efficicency
    """
    # pylint: disable-msg=protected-access
    _z_src = np.array(z_src)
    return (
        np.heaviside(_z_src - z_cl, 0) * cosmo._eval_da_z1z2(z_cl, _z_src) / cosmo._eval_da(_z_src)
    )


def compute_beta_s(z_src, z_cl, z_inf, cosmo):
    r"""Geometric lensing efficicency ratio

    .. math::
        \beta_s = \beta(z_{src})/\beta(z_{inf})

    Parameters
    ----------
    z_src : float, array_like
        Source galaxy redshift
    z_cl: float
        Galaxy cluster redshift
    z_inf: float
        Redshift at infinity
    cosmo: clmm.Cosmology
        CLMM Cosmology object

    Returns
    -------
    numpy array
        Geometric lensing efficicency ratio
    """
    beta_s = compute_beta(z_src, z_cl, cosmo) / compute_beta(z_inf, z_cl, cosmo)
    return beta_s


def compute_beta_s_func(z_src, z_cl, z_inf, cosmo, func, *args, **kwargs):
    r"""Geometric lensing efficicency ratio times a value of a function

    .. math::
        \beta_{s}\times \text{func} = \beta_s(z_{src}, z_{cl}, z_{inf})
        \times\text{func}(*args,\ **kwargs)

    Parameters
    ----------
    z_src : array_like, float, function
        Information on the background source galaxy redshift(s). Value required depends on
        `z_src_info` (see below).
    z_cl: float
        Galaxy cluster redshift
    z_inf: float
        Redshift at infinity
    cosmo: clmm.Cosmology
        CLMM Cosmology object
    func: callable
        A scalar function
    *args: positional arguments
        args to be passed to `func`
    **kwargs: keyword arguments
        kwargs to be passed to `func`

    Returns
    -------
    numpy array
        Geometric lensing efficicency ratio for each source
    """
    beta_s = compute_beta(z_src, z_cl, cosmo) / compute_beta(z_inf, z_cl, cosmo)
    beta_s_func = beta_s * func(*args, **kwargs)
    return beta_s_func


def compute_beta_s_mean_from_distribution(
    z_cl, z_inf, cosmo, zmax=10.0, delta_z_cut=0.1, zmin=None, z_distrib_func=None
):
    r"""Mean value of the geometric lensing efficicency

    .. math::
       \left<\beta_s\right> = \frac{\int_{z = z_{min}}^{z_{max}}\beta_s(z)N(z)}
       {\int_{z = z_{min}}^{z_{max}}N(z)}

    Parameters
    ----------
    z_cl: float
        Galaxy cluster redshift
    z_inf: float
        Redshift at infinity
    cosmo: clmm.Cosmology
        CLMM Cosmology object
    zmax: float, optional
        Maximum redshift to be set as the source of the galaxy when performing the sum.
        Default: 10
    delta_z_cut: float, optional
        Redshift interval to be summed with :math:`z_{cl}` to return :math:`z_{min}`.
        This feature is not used if :math:`z_{min}` is provided by the user. Default: 0.1
    zmin: float, None, optional
        Minimum redshift to be set as the source of the galaxy when performing the sum.
        Default: None
    z_distrib_func: one-parameter function, optional
        Redshift distribution function. Default is Chang et al (2013) distribution function.

    Returns
    -------
    float
        Mean value of the geometric lensing efficicency
    """
    if z_distrib_func is None:
        z_distrib_func = zdist.chang2013

    def integrand(z_i):
        return compute_beta_s(z_i, z_cl, z_inf, cosmo) * z_distrib_func(z_i)

    if zmin is None:
        zmin = z_cl + delta_z_cut

    return quad(integrand, zmin, zmax)[0] / quad(z_distrib_func, zmin, zmax)[0]


def compute_beta_s_square_mean_from_distribution(
    z_cl, z_inf, cosmo, zmax=10.0, delta_z_cut=0.1, zmin=None, z_distrib_func=None
):
    r"""Mean square value of the geometric lensing efficicency ratio

    .. math::
       \left<\beta_s^2\right> =\frac{\int_{z = z_{min}}^{z_{max}}\beta_s^2(z)N(z)}
       {\int_{z = z_{min}}^{z_{max}}N(z)}

    Parameters
    ----------
    z_cl: float
        Galaxy cluster redshift
    z_inf: float
        Redshift at infinity
    cosmo: clmm.Cosmology
        CLMM Cosmology object
    zmax: float
        Minimum redshift to be set as the source of the galaxy\
        when performing the sum.
    delta_z_cut: float
        Redshift interval to be summed with $z_cl$ to return\
        $zmin$. This feature is not used if $z_min$ is provided by the user.
    zmin: float, None, optional
        Minimum redshift to be set as the source of the galaxy when performing the sum.
        Default: None
    z_distrib_func: one-parameter function, optional
        Redshift distribution function. Default is Chang et al (2013) distribution function.
    Returns
    -------
    float
        Mean square value of the geometric lensing efficicency ratio.
    """
    if z_distrib_func is None:
        z_distrib_func = zdist.chang2013

    def integrand(z_i):
        return compute_beta_s(z_i, z_cl, z_inf, cosmo) ** 2 * z_distrib_func(z_i)

    if zmin is None:
        zmin = z_cl + delta_z_cut

    return quad(integrand, zmin, zmax)[0] / quad(z_distrib_func, zmin, zmax)[0]


def compute_beta_s_mean_from_weights(z_src, z_cl, z_inf, cosmo, shape_weights):
    r"""Mean square value of the geometric lensing efficicency ratio

    .. math::
       \left<\beta_s\right> =\frac{\sum_i \beta_s(z_i)w_i}
       {\sum_i w_i}

    Parameters
    ----------
    z_src: float, array_like
        Invididual source galaxies redshift.
    z_cl: float
        Galaxy cluster redshift.
    z_inf: float
        Redshift at infinity.
    cosmo: clmm.Cosmology
        CLMM Cosmology object
    shape_weights: float, array_like
        Individual source galaxies shape weights.\
        If not None, the function uses Eq.(13) from\
        https://arxiv.org/pdf/1611.03866.pdf with evenly distributed\
        weights summing to one.

    Returns
    -------
    float
        Mean value of the geometric lensing efficicency ratio.
    """
    _z_src = np.array(z_src)
    if shape_weights is None:
        _shape_weights = np.ones_like(_z_src)
    else:
        _shape_weights = np.array(shape_weights)
    beta_s = compute_beta_s(_z_src, z_cl, z_inf, cosmo)
    return (_shape_weights * beta_s).sum() / _shape_weights.sum()


def compute_beta_s_square_mean_from_weights(
    z_src,
    z_cl,
    z_inf,
    cosmo,
    shape_weights,
):
    r"""Mean square value of the geometric lensing efficicency ratio

    .. math::
       \left<\beta_s^2\right> =\frac{\sum_i \beta_s^2(z_i)w_i}
       {\sum_i w_i}

    Parameters
    ----------
    z_src: float, array_like
        Invididual source galaxies redshift.
    z_cl: float
        Galaxy cluster redshift.
    z_inf: float
        Redshift at infinity.
    cosmo: clmm.Cosmology
        CLMM Cosmology object
    shape_weights: float, array_like
        Individual source galaxies shape weights.
    Returns
    -------
    float
        Mean square value of the geometric lensing efficicency ratio.
    """
    _z_src = np.array(z_src)
    if shape_weights is None:
        _shape_weights = np.ones_like(_z_src)
    else:
        _shape_weights = np.array(shape_weights)
    beta_s = compute_beta_s(_z_src, z_cl, z_inf, cosmo)
    return (_shape_weights * beta_s**2).sum() / _shape_weights.sum()



"""General utility functions that are used in multiple modules"""

import warnings

import numpy as np
import qp
from scipy.integrate import simpson


def _integ_pzfuncs(pzpdf, pzbins, zmin=0.0, zmax=5, kernel=None, ngrid=1000):
    r"""
    Integrates the product of a photo-z pdf with a given kernel.
    This function was created to allow for data with different photo-z binnings.

    Parameters
    ----------
    pzpdf : list of arrays
        Photometric probablility density functions of the source galaxies.
    pzbins : list of arrays
        Redshift axis on which the individual photoz pdf is tabulated.
    zmin : float, optional
        Minimum redshift for integration. Default: 0
    zmax : float, optional
        Maximum redshift for integration. Default: 5
    kernel : function, optional
        Function to be integrated with the pdf, must be f(z_array) format.
        Default: kernel(z)=1
    ngrid : int, optional
        Number of points for the interpolation of the redshift pdf.

    Returns
    -------
    numpy.ndarray
        Kernel integrated with the pdf of each galaxy.
    """

    if len(np.array(pzbins).shape) > 1:
        # Each galaxy as a diiferent zbin
        _qp_type = qp.interp_irregular
    else:
        _qp_type = qp.interp

    qp_ensemble = qp.Ensemble(
        _qp_type,
        data={
            "xvals": np.array(pzbins),
            "yvals": np.array(pzpdf),
        },
    )

    if kernel is None:

        def kernel(z):
            # pylint: disable=unused-argument
            return 1.0

    z_grid = np.linspace(zmin, zmax, ngrid)
    pz_matrix = qp_ensemble.pdf(z_grid)
    kernel_matrix = kernel(z_grid)

    return simpson(pz_matrix * kernel_matrix, x=z_grid, axis=1)


def compute_for_good_redshifts(
    function,
    z1,
    z2,
    bad_value,
    warning_message,
    z1_arg_name="z1",
    z2_arg_name="z2",
    r_proj=None,
    **kwargs,
):
    """Computes function only for `z1` < `z2`, the rest is filled with `bad_value`

    Parameters
    ----------
    function: function
        Function to be executed
    z1: float, array_like
        Redshift lower
    z2: float, array_like
        Redshift higher
    bad_value: any
        Value to fill when `z1` >= `z2`
    warning_message: str
        Warning message to be displayed when `z1` >= `z2`
    z1_arg_name: str, optional
        Name of the keyword argument that `z1` is passed to. Default: 'z1'
    z2_arg_name: str, optional
        Name of the keyword argument that `z2` is passed to. Default: 'z2'
    r_proj: float, array_like, optional
        Value to be passed to keyword argument `r_proj` of `function`. Default: None

    Returns
    -------
    Return type of `function`
        Output of `function` with value for `z1` >= `z2` replaced by `bad_value`
    """
    kwargs = {z1_arg_name: locals()["z1"], z2_arg_name: locals()["z2"], **kwargs}

    z_good = np.less(z1, z2)
    if r_proj is not None:
        r_proj = np.array(r_proj) * np.full_like(z_good, True)
        z_good = z_good * r_proj.astype(bool)
        kwargs.update({"r_proj": r_proj[z_good] if np.iterable(r_proj) else r_proj})

    if not np.all(z_good):
        warnings.warn(warning_message, stacklevel=2)
        if np.iterable(z_good):
            res = np.full(z_good.shape, bad_value)
            if np.any(z_good):
                kwargs[z1_arg_name] = np.array(z1)[z_good] if np.iterable(z1) else z1
                kwargs[z2_arg_name] = np.array(z2)[z_good] if np.iterable(z2) else z2
                res[z_good] = function(**kwargs)
        else:
            res = bad_value
    else:
        res = function(**kwargs)
    return res



"""@file.py sampler.py
Functions for sampling (output either peak or full distribution)
"""

from scipy import optimize as spo


def sciopt(model_to_shear_profile, logm_0, **kwargs):
    r"""Uses scipy optimize minimize to output the peak

    Parameters
    ----------
    model_to_shear_profile : callable
        The objective function to be minimized.
        ``model_to_shear_profile(x, *args) -> float``
        where ``x`` is an 1-D array with shape (n,) and ``args``
        is a tuple of the fixed parameters needed to completely
        specify the function.
    logm_0 : ndarray, shape (n,)
        Initial guess. Array of real elements of size (n,),
        where 'n' is the number of independent variables.
    kwargs :
        Other optional keyword arguments passed to `minimize`.
        Please check the `scipy` documentaion for information on default values and methods.

    Returns
    -------
    x : array
    The solution of the optimization

    """
    return spo.minimize(model_to_shear_profile, logm_0, **kwargs).x


def basinhopping(model_to_shear_profile, logm_0, **kwargs):
    r"""Uses basinhopping, a scipy global optimization function, to find the minimum.

    Parameters
    ----------
    model_to_shear_profile : callable
        The objective function to be minimized.
        ``model_to_shear_profile(x, *args) -> float``
        where ``x`` is an 1-D array with shape (n,) and ``args``
        is a tuple of the fixed parameters needed to completely
        specify the function.
    logm_0 : array_like
        Initial guess. Array of real elements of size (n,),
        where 'n' is the number of independent variables.
    kwargs :
        Other optional keyword arguments passed to `basinhopping`.
        Please check the `scipy` documentaion for information on default values and methods.

    Returns
    -------
    x : array
    The solution of the optimization

    """
    return spo.basinhopping(model_to_shear_profile, logm_0, **kwargs).x


def scicurve_fit(profile_model, radius, profile, err_profile, absolute_sigma=True, **kwargs):
    r"""
    Uses scipy.optimize.curve_fit to find best fit parameters

    Parameters
    ----------
    profile_model : callable
        The model function, f(x, ...). It must take the independent
        variable as the first argument and the parameters to fit as
        separate remaining arguments.
    radius : array_like or object
        The independent variable where the data is measured.
        Should usually be an M-length sequence or an (k,M)-shaped array for
        functions with k predictors, but can actually be any object.
    profile : array_like
        The dependent data, a length M array - nominally ``f(xdata, ...)``.
    err_profile : M-length sequence or MxM array
        Determines the uncertainty in `ydata`. If we define residuals as
        ``r = ydata - f(xdata, *popt)``, then the interpretation of `sigma`
        depends on its number of dimensions:
        - A 1-D `sigma` should contain values of standard deviations of
        errors in `ydata`. In this case, the optimized function is
        ``chisq = sum((r / sigma) ** 2)``.
        - A 2-D `sigma` should contain the covariance matrix of
        errors in `ydata`. In this case, the optimized function is
        ``chisq = r.T @ inv(sigma) @ r``.
    absolute_sigma : bool, optional
        If True (default), `sigma` is used in an absolute sense and the estimated parameter
        covariance `pcov` reflects these absolute values.
        If False, only the relative magnitudes of the `sigma` values matter.
        The returned parameter covariance matrix `pcov` is based on scaling
        `sigma` by a constant factor. This constant is set by demanding that the
        reduced `chisq` for the optimal parameters `popt` when using the
        *scaled* `sigma` equals unity. In other words, `sigma` is scaled to
        match the sample variance of the residuals after the fit. Default is True.
        Mathematically,
        ``pcov(absolute_sigma=False) = pcov(absolute_sigma=True) * chisq(popt)/(M-N)``

    kwargs :
        Other optional keyword arguments passed to `curve_fit`
        Please check the `scipy` documentaion for information on default values and methods.

    Returns
    -------
    p : list
        contains :
            p[0] : array
                Optimal values for the parameters so that the sum of the squared
                residuals of ``f(xdata, *popt) - ydata`` is minimized.
            p[1] : 2-D array
                The estimated covariance of popt. The diagonals provide the variance
                of the parameter estimate. To compute one standard deviation errors
                on the parameters use ``perr = np.sqrt(np.diag(pcov))``.
                How the `sigma` parameter affects the estimated covariance
                depends on `absolute_sigma` argument, as described above.

    """
    return spo.curve_fit(
        profile_model, radius, profile, sigma=err_profile, absolute_sigma=absolute_sigma, **kwargs
    )


samplers = {"minimize": sciopt, "basinhopping": basinhopping}

fitters = {
    "curve_fit": scicurve_fit,
}



"""@file numcosmo.py
NumCosmo implementation of CLMModeling
"""

import math

import gi
import numpy as np
from gi.repository import NumCosmo as Nc
from gi.repository import NumCosmoMath as Ncm

from ..cosmology.numcosmo import NumCosmoCosmology
from .parent_class import CLMModeling

gi.require_version("NumCosmo", "1.0")
gi.require_version("NumCosmoMath", "1.0")


class NumCosmoCLMModeling(CLMModeling):
    r"""Object with functions for halo mass modeling

    Attributes
    ----------
    backend: str
        Name of the backend being used
    massdef : str
        Profile mass definition ("mean", "critical", "virial" - letter case independent)
    delta_mdef : int
        Mass overdensity definition.
    halo_profile_model : str
        Profile model parameterization ("nfw", "einasto", "hernquist" - letter case independent)
    cosmo: Cosmology
        Cosmology object
    hdpm: Object
        Backend object with halo profiles
    mdef_dict: dict
        Dictionary with the definitions for mass
    hdpm_dict: dict
        Dictionary with the definitions for profile
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=abstract-method

    def __init__(
        self,
        massdef="mean",
        delta_mdef=200,
        halo_profile_model="nfw",
        validate_input=True,
    ):
        CLMModeling.__init__(self, validate_input)
        # Update class attributes
        Ncm.cfg_init()
        self.backend = "nc"
        self.mdef_dict = {
            "mean": Nc.HaloDensityProfileMassDef.MEAN,
            "critical": Nc.HaloDensityProfileMassDef.CRITICAL,
            "virial": Nc.HaloDensityProfileMassDef.VIRIAL,
        }
        self.hdpm_dict = {
            "nfw": Nc.HaloDensityProfileNFW.new,
            "einasto": Nc.HaloDensityProfileEinasto.new,
            "hernquist": Nc.HaloDensityProfileHernquist.new,
        }
        self.cosmo_class = NumCosmoCosmology
        # Set halo profile and cosmology
        self.set_halo_density_profile(halo_profile_model, massdef, delta_mdef)
        self.set_cosmo(None)

    # Functions implemented by child class

    def _update_halo_density_profile(self):
        """updates halo density profile with set internal properties"""
        # Makes sure current cdelta/mdelta values are kept
        has_cm_vals = self.hdpm is not None
        if has_cm_vals:
            cdelta = self.cdelta
            log10_mdelta = self.hdpm.props.log10MDelta

        self.hdpm = self.hdpm_dict[self.halo_profile_model](
            self.mdef_dict[self.massdef], self.delta_mdef
        )

        if has_cm_vals:
            self.cdelta = cdelta
            self.hdpm.props.log10MDelta = log10_mdelta

        self._update_vec_funcs()

    def _get_concentration(self):
        """get concentration"""
        return self.hdpm.props.cDelta

    def _get_mass(self):
        """get mass"""
        return 10**self.hdpm.props.log10MDelta

    def _set_concentration(self, cdelta):
        """set concentration"""
        self.hdpm.props.cDelta = cdelta
        self._update_vec_funcs()

    def _set_mass(self, mdelta):
        """set mass"""
        self.hdpm.props.log10MDelta = math.log10(mdelta)
        self._update_vec_funcs()

    def _set_einasto_alpha(self, alpha):
        if alpha is None:
            self.hdpm.props.alpha = 0.25
        else:
            self.hdpm.props.alpha = alpha
        self._update_vec_funcs()

    def _get_einasto_alpha(self, z_cl=None):
        """get the value of the Einasto slope"""
        # Note that z_cl is needed for CCL<2.6 only
        return self.hdpm.props.alpha

    def _get_delta_mdef_virial(self, z_cl):
        "Gets the virial overdensity"
        # This fuction can return the overdensity for all mass definitions,
        # but in CLMM, we only use it in the case of massdef='virial'.
        return self.hdpm.Delta(self.cosmo.be_cosmo, z_cl)

    def _eval_reduced_tangential_shear_core(self, r_proj, z_cl, z_src):
        """eval reduced tangential shear considering a single redshift plane
        for background sources"""

        if (
            isinstance(r_proj, (list, np.ndarray))
            and isinstance(z_src, (list, np.ndarray))
            and len(r_proj) == len(z_src)
        ):
            func = self.cosmo.smd.reduced_shear_array_equal
        else:
            func = self.cosmo.smd.reduced_shear_array
        return func(
            self.hdpm,
            self.cosmo.be_cosmo,
            np.atleast_1d(r_proj),
            1.0,
            1.0,
            np.atleast_1d(z_src),
            z_cl,
            z_cl,
        )

    # Functions unique to this class

    def _update_vec_funcs(self):
        """Set/update all functions that are vectorized"""
        self._eval_3d_density = np.vectorize(
            lambda r3d, z_cl: self.hdpm.eval_density(self.cosmo.be_cosmo, r3d, z_cl)
        )
        self._eval_surface_density = np.vectorize(
            lambda r_proj, z_cl: self.cosmo.smd.sigma(self.hdpm, self.cosmo.be_cosmo, r_proj, z_cl)
        )
        self._eval_mean_surface_density = np.vectorize(
            lambda r_proj, z_cl: self.cosmo.smd.sigma_mean(
                self.hdpm, self.cosmo.be_cosmo, r_proj, z_cl
            )
        )
        self._eval_excess_surface_density = np.vectorize(
            lambda r_proj, z_cl: self.cosmo.smd.sigma_excess(
                self.hdpm, self.cosmo.be_cosmo, r_proj, z_cl
            )
        )
        self._eval_tangential_shear_core = np.vectorize(
            lambda r_proj, z_cl, z_src: self.cosmo.smd.shear(
                self.hdpm, self.cosmo.be_cosmo, r_proj, z_src, z_cl, z_cl
            )
        )
        self._eval_convergence_core = np.vectorize(
            lambda r_proj, z_cl, z_src: self.cosmo.smd.convergence(
                self.hdpm, self.cosmo.be_cosmo, r_proj, z_src, z_cl, z_cl
            )
        )
        self._eval_magnification_core = np.vectorize(
            lambda r_proj, z_cl, z_src: self.cosmo.smd.magnification(
                self.hdpm, self.cosmo.be_cosmo, r_proj, z_src, z_cl, z_cl
            )
        )

    def get_mset(self):
        r"""
        Gets a mass set (NumCosmo internal use)
        """
        mset = Ncm.MSet.empty_new()
        mset.set(self.cosmo.be_cosmo)
        mset.set(self.hdpm)
        mset.set(self.cosmo.smd)
        return mset

    def set_mset(self, mset):
        r"""
        Sets a mass set (NumCosmo internal use)
        """
        self.cosmo.set_be_cosmo(mset.get(Nc.HICosmo.id()))

        self.hdpm = mset.get(Nc.HaloDensityProfile.id())
        self.cosmo.smd = mset.get(Nc.WLSurfaceMassDensity.id())
        self.cosmo.smd.prepare_if_needed(self.cosmo.be_cosmo)


Cosmology = NumCosmoCosmology
Modeling = NumCosmoCLMModeling



"""@file numcosmo.py
Cosmology using NumCosmo
"""

import gi
import numpy as np

# pylint: disable=wrong-import-position
gi.require_version("NumCosmo", "1.0")
gi.require_version("NumCosmoMath", "1.0")
from gi.repository import NumCosmo as Nc
from gi.repository import NumCosmoMath as Ncm

from .parent_class import CLMMCosmology

__all__ = []


class NumCosmoCosmology(CLMMCosmology):
    """
    Cosmology object

    Attributes
    ----------
    backend: str
        Name of back-end used
    be_cosmo: cosmology library
        Cosmology library used in the back-end
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=abstract-method

    def __init__(self, dist=None, dist_zmax=15.0, **kwargs):
        self.dist = None
        self.smd = None

        super().__init__(**kwargs)

        # this tag will be used to check if the cosmology object is accepted by the modeling
        self.backend = "nc"

        # set kmin/kmax for powerspectrum computations
        self.additional_config["pk_kmin"] = 1.0e-5
        self.additional_config["pk_kmax"] = 1.0
        self.additional_config["ns"] = 0.96
        self.additional_config["sigma8"] = 0.8
        self.additional_config["sigma8_eps"] = 1e-8

        if dist:
            self.set_dist(dist)
        else:
            self.set_dist(Nc.Distance.new(dist_zmax))

        self.smd = Nc.WLSurfaceMassDensity.new(self.dist)
        self.smd.prepare_if_needed(self.be_cosmo)

    def _init_from_cosmo(self, be_cosmo):
        assert isinstance(be_cosmo, Nc.HICosmo)
        assert isinstance(be_cosmo, Nc.HICosmoDECpl)
        assert isinstance(be_cosmo.peek_reparam(), Nc.HICosmoDEReparamOk)
        self.be_cosmo = be_cosmo
        self._update_vec_funcs()

    def _init_from_params(self, H0, Omega_b0, Omega_dm0, Omega_k0):
        # pylint: disable=arguments-differ
        self.be_cosmo = Nc.HICosmoDECpl(massnu_length=1)
        self.be_cosmo.omega_x2omega_k()
        self.be_cosmo.param_set_by_name("w0", -1.0)
        self.be_cosmo.param_set_by_name("w1", 0.0)
        self.be_cosmo.param_set_by_name("Tgamma0", 2.7255)
        self.be_cosmo.param_set_by_name("massnu_0", 0.06)
        self.be_cosmo.param_set_by_name("H0", H0)
        self.be_cosmo.param_set_by_name("Omegab", Omega_b0)
        self.be_cosmo.param_set_by_name("Omegac", Omega_dm0)
        self.be_cosmo.param_set_by_name("Omegak", Omega_k0)

        cosmo = self.be_cosmo
        en_nu = 3.046 - 3.0 * cosmo.E2Press_mnu(1.0e10) / (
            cosmo.E2Omega_g(1.0e10) * (7.0 / 8.0 * (4.0 / 11.0) ** (4.0 / 3.0))
        )

        self.be_cosmo.param_set_by_name("ENnu", en_nu)
        self._update_vec_funcs()

    def _set_param(self, key, value):
        if key == "Omega_b0":
            self.be_cosmo.param_set_by_name("Omegab", value)
        elif key == "Omega_dm0":
            self.be_cosmo.param_set_by_name("Omegac", value)
        elif key == "Omega_k0":
            self.be_cosmo.param_set_by_name("Omegak", value)
        elif key == "h":
            self.be_cosmo.param_set_by_name("H0", value * 100.0)
        elif key == "H0":
            self.be_cosmo.param_set_by_name("H0", value)
        else:
            raise ValueError(f"Unsupported parameter {key}")
        self._update_vec_funcs()

    def _get_param(self, key):
        if key == "Omega_m0":
            value = self.be_cosmo.Omega_m0()
        elif key == "Omega_b0":
            value = self.be_cosmo.Omega_b0()
        elif key == "Omega_dm0":
            value = self.be_cosmo.Omega_c0()
        elif key == "Omega_k0":
            value = self.be_cosmo.Omega_k0()
        elif key == "h":
            value = self.be_cosmo.h()
        elif key == "H0":
            value = self.be_cosmo.H0()
        else:
            raise ValueError(f"Unsupported parameter {key}")
        return value

    def set_dist(self, dist):
        r"""Sets distance functions (NumCosmo internal use)"""
        assert isinstance(dist, Nc.Distance)
        self.dist = dist
        self.dist.prepare_if_needed(self.be_cosmo)

    def _get_Omega_m(self, z):
        return self._get_E2Omega_m(z) / self._get_E2(z)

    def _update_vec_funcs(self):
        """Update all functions that are vectorized"""

        self._get_E2 = np.vectorize(self.be_cosmo.E2)
        self._get_E2Omega_m = np.vectorize(self.be_cosmo.E2Omega_m)
        self._eval_da_z1z2_core = np.vectorize(
            lambda z1, z2: (
                self.dist.angular_diameter_z1_z2(self.be_cosmo, z1, z2) * self.be_cosmo.RH_Mpc()
            )
        )
        self._eval_sigma_crit_core = np.vectorize(
            lambda z_len, z_src: (self.smd.sigma_critical(self.be_cosmo, z_src, z_len, z_len))
        )
        if self.dist is not None:
            self.dist.prepare_if_needed(self.be_cosmo)
        if self.smd is not None:
            self.smd.prepare_if_needed(self.be_cosmo)

    def _get_rho_c(self, z):
        return (
            Ncm.C.crit_mass_density_h2_solar_mass_Mpc3()
            * self._get_param("h") ** 2
            * self._get_E2(z)
        )

    def _get_rho_m(self, z):
        # total matter density in physical units [Msun/Mpc3]
        rho_m = (
            self._get_E2Omega_m(z)
            * Ncm.C.crit_mass_density_h2_solar_mass_Mpc3()
            * self._get_param("h")
            * self._get_param("h")
        )
        return rho_m

    def _eval_linear_matter_powerspectrum(self, k_vals, redshift):
        # Using the EH transfer function as this is the
        # default for the CCL backend as well
        pow_spec = Nc.PowspecMLTransfer.new(Nc.TransferFuncEH.new())
        pow_spec.set_kmin(self.additional_config["pk_kmin"])
        pow_spec.set_kmax(self.additional_config["pk_kmax"])

        # Instead, computing the PS from the CLASS backend of Numcosmo
        # pow_spec  = Nc.PowspecMLCBE.new ()
        # pow_spec.peek_cbe().props.use_ppf = True

        if self.be_cosmo.reion is None:
            reion = Nc.HIReionCamb.new()
            self.be_cosmo.add_submodel(reion)
        if self.be_cosmo.prim is None:
            prim = Nc.HIPrimPowerLaw.new()
            self.be_cosmo.add_submodel(prim)
            # The default CLMM cosmology has ns=0.96 and sigma8=0.8
            # Need to adapt the NC cosmology accordingly
            self.be_cosmo.prim.props.n_SA = self.additional_config["ns"]
            sigma8 = pow_spec.sigma_tophat_R(  # computes sigma8 at z=0
                self.be_cosmo,
                self.additional_config["sigma8_eps"],
                0.0,
                8.0 / self.be_cosmo.h(),
            )
            old_amplitude = np.exp(self.be_cosmo.prim.props.ln10e10ASA)
            self.be_cosmo.prim.props.ln10e10ASA = np.log(
                (self.additional_config["sigma8"] / sigma8) ** 2 * old_amplitude
            )

        pow_spec.prepare(self.be_cosmo)

        return [pow_spec.eval(self.be_cosmo, redshift, k) for k in k_vals]



"""Functions to generate mock source galaxy distributions to demo lensing code"""

import warnings

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from scipy.special import erfc

from ..gcdata import GCData
from ..theory import compute_convergence, compute_tangential_shear
from ..utils import (
    _draw_random_points_from_distribution,
    _validate_coordinate_system,
    compute_lensed_ellipticity,
    convert_units,
    gaussian,
    redshift_distributions,
    validate_argument,
)


def generate_galaxy_catalog(
    cluster_m,
    cluster_z,
    cluster_c,
    cosmo,
    zsrc,
    cluster_ra=0.0,
    cluster_dec=0.0,
    delta_so=200,
    massdef="mean",
    halo_profile_model="nfw",
    zsrc_min=None,
    zsrc_max=7.0,
    field_size=8.0,
    shapenoise=None,
    mean_e_err=None,
    photoz_sigma_unscaled=None,
    nretry=5,
    ngals=None,
    ngal_density=None,
    pz_bins=101,
    pz_quantiles_conf=(5, 31),
    pzpdf_type="shared_bins",
    coordinate_system=None,
    validate_input=True,
):
    r"""Generates a mock dataset of sheared background galaxies.

    We build galaxy catalogs following a series of steps.

    1. Draw true redshifts of the source galaxy population. This step is described by the
    parameters `zsrc` and `zsrc_max`. `zsrc` can be a `float` in which case every source is
    at the given redshift or a `str` describing a specific model to use for the source
    distribution. Currently, the only supported model for source galaxy distribution is that
    of Chang et al. 2013 arXiv:1305.0793. When a model is used to describe the distribution,
    `zsrc_max` is the maximum allowed redshift of a source galaxy.

    2. Apply photometric redshift errors to the source galaxy population. This step is
    described by the parameter `photoz_sigma_unscaled`. If this parameter is set to a float,
    we add Gaussian uncertainty to the source redshift

    .. math::
        z \sim \mathcal{N}\left(z^{\rm true},
        \sigma_{\rm photo-z}^{\rm unscaled}(1+z^{\rm true}) \right)

    We additionally include two columns in the output catalog, `pzbins` and `pzpdf` which
    desribe the photo-z distribution as a Gaussian centered at :math:`z^{\rm true}` with a
    width :math:`\sigma_{\rm photo-z} = \sigma_{\rm photo-z}^{\rm unscaled}(1+z^{\rm true})`

    If `photoz_sigma_unscaled` is `None`, the `z` column in the output catalog is the true
    redshift.

    3. Draw galaxy positions. Positions are drawn in a square box around the lens position (with a
    default side length of 8 Mpc) at the lens redhsift. We then convert to right ascension and
    declination using the cosmology defined in `cosmo`.

    4. We predict the reduced tangential shear of each using the radial distances of each source
    from the lens, the source redshifts, and the lens mass, concentration, and redshift. In the
    given cosmology for an NFW halo. The reduced tangential shear is then transformed into `g1`
    and `g2` components.

    5. If the `shapenoise=True`, intrinsic ellipticities (1,2) components are drawn from a
    Gaussian distribution of width of `shapenoise`. These ellipticities components are then
    combined with `g1` and `g2` to provide lensed ellipticies `e1` and `e2`. If
    `shapenoise=False`, `g1` and `g2` are directly used as ellipticity components.


    If the shape noise parameter is high, we may draw nonsensical values for ellipticities. We
    ensure that we does not return any nonsensical values for derived properties. We re-draw
    all galaxies with e1 or e2 outside the bounds of [-1, 1]. After 5 (default) attempts to
    re-draw these properties, we return the catalog as is and throw a warning.

    Parameters
    ----------
    cluster_m : float
        Cluster mass in Msun
    cluster_z : float
        Cluster redshift
    cluster_c : float
        Cluster concentration in the same mass definition as delta_so
    cosmo: clmm.Cosmology, optional
        Cosmology object.
    zsrc : float or str
        Choose the source galaxy distribution to be fixed or drawn from a predefined distribution.

        * `float` : All sources galaxies at this fixed redshift;
        * `str` : Draws individual source gal redshifts from predefined distribution. Options are:

            * `chang13` - Chang et al. 2013 (arXiv:1305.0793);
            * `desc_srd` - LSST/DESC Science Requirement Document (arxiv:1809.01669);

    delta_so : float, optional
        Overdensity density contrast used to compute the cluster mass and concentration. The
        spherical overdensity mass is computed as the mass enclosed within the radius
        :math:`R_{\Delta{\rm SO}}` where the mean matter density is :math:`\Delta_{\rm SO}` times
        the mean (or critical, depending on the massdef keyword) density of the Universe at the
        cluster redshift
        :math:`M_{\Delta{\rm SO}}=4/3\pi\Delta_{\rm SO}\rho_{m}(z_{\rm lens})R_{\Delta{\rm SO}}^3`
    massdef : string, optional
        Definition the mass overdensity with respect to the 'mean' or 'critical' density of the
        universe. Default is 'mean' as it works for all theory backends. The NumCosmo and CCL
        backends also allow the use of 'critical'.  (letter case independent)
    halo_profile_model : string, optional
        Halo density profile. Default is 'nfw', which works for all theory backends. The NumCosmo
        backend allow for more options, e.g. 'einasto' or 'burkert' profiles (letter case
        independent).
    zsrc_min : float, optional
        The minimum true redshift of the sources. If photoz errors are included, the observed
        redshift may be smaller than zsrc_min.
    zsrc_max : float, optional
        The maximum true redshift of the sources, apllied when galaxy redshifts are drawn from a
        redshift distribution. If photoz errors are included, the observed redshift may be larger
        than zsrc_max.
    field_size : float, optional
        The size of the field (field_size x field_size) to be simulated.
        Proper distance in Mpc at the cluster redshift.
    shapenoise : float, optional
        If set, applies Gaussian shape noise to the galaxy shapes with a width set by `shapenoise`
    mean_e_err : float, optional
        Mean per-component ellipticity uncertainty. Currently,
        individual uncertainties are drawn from a uniform distribution
        in the range [0.9,1.1]*mean_e_err. If not provided, the output
        table will not include this column.
    photoz_sigma_unscaled : float, optional
        If set, applies photo-z errors to source redshifts
    pz_bins: int, array
        Photo-z pdf bins in the given range. If int, the limits are set automatically.
        If is array, must be the bin edges.
    pz_quantiles_conf: tuple
        Configuration for quantiles when `pzpdf_type='quantiles'`. Must be with the format
        `(max_sigma_dev, num_points)`, which is used as
        `sigma_steps = np.linspace(-max_sigma_dev, max_sigma_dev, num_points)`
    pzpdf_type: str, None
        Type of photo-z pdf to be stored, options are:
            `None` - does not store PDFs;
            `'shared_bins'` - single binning for all galaxies
            `'individual_bins'` - individual binning for each galaxy
            `'quantiles'` - quantiles of PDF
    nretry : int, optional
        The number of times that we re-draw each galaxy with non-sensical derived properties
    ngals : float, optional
        Number of galaxies to generate
    ngal_density : float, optional
        The number density of galaxies (in galaxies per square arcminute, from z=0 to z=infty).
        The number of galaxies to be drawn will then depend on the redshift distribution and
        user-defined redshift range. If specified, the ngals argument will be ignored.
    coordinate_system : str, optional
        Coordinate system of the ellipticity components. Must be either 'celestial' or
        euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        If not set, defaults to 'euclidean'.
    validate_input: bool
        Validade each input argument

    Returns
    -------
    galaxy_catalog : clmm.GCData
        Table of source galaxies with drawn and derived properties required for lensing studies

    Notes
    -----
    Much of this code in this function was adapted from the Dallas group
    """
    # pylint: disable=too-many-arguments
    # Too many local variables (25/15)
    # pylint: disable=R0914

    if not coordinate_system:
        warnings.warn("coordinate_system not set, defaulting to 'euclidean'")
        coordinate_system = "euclidean"

    if validate_input:
        validate_argument(locals(), "cluster_m", float, argmin=0, eqmin=True)
        validate_argument(locals(), "cluster_z", float, argmin=0, eqmin=True)
        validate_argument(locals(), "cluster_c", float, argmin=0, eqmin=True)
        validate_argument(
            locals(),
            "cluster_dec",
            float,
            argmin=-90,
            argmax=90,
            eqmin=True,
            eqmax=True,
        )
        validate_argument(
            locals(),
            "cluster_ra",
            float,
            argmin=-360.0,
            argmax=360.0,
            eqmin=True,
            eqmax=True,
        )
        validate_argument(locals(), "zsrc", (float, str))
        validate_argument(locals(), "delta_so", float, argmin=0, eqmin=True)
        validate_argument(locals(), "massdef", str)
        validate_argument(locals(), "halo_profile_model", str)
        validate_argument(locals(), "zsrc_min", float, argmin=0, eqmin=True, none_ok=True)
        validate_argument(locals(), "zsrc_max", float, argmin=0, eqmin=True)
        validate_argument(locals(), "field_size", float, argmin=0, eqmin=True)
        validate_argument(locals(), "shapenoise", float, argmin=0, none_ok=True)
        validate_argument(locals(), "mean_e_err", float, argmin=0, none_ok=True)
        validate_argument(locals(), "photoz_sigma_unscaled", float, argmin=0, none_ok=True)
        validate_argument(locals(), "nretry", int)
        validate_argument(locals(), "ngals", float, none_ok=True)
        validate_argument(locals(), "ngal_density", float, none_ok=True)
        validate_argument(locals(), "pz_bins", (int, "array"))
        _validate_coordinate_system(locals(), "coordinate_system")

    if zsrc_min is None:
        zsrc_min = cluster_z + 0.1

    params = {
        "cluster_m": cluster_m,
        "cluster_z": cluster_z,
        "cluster_c": cluster_c,
        "cluster_ra": cluster_ra,
        "cluster_dec": cluster_dec,
        "cosmo": cosmo,
        "delta_so": delta_so,
        "zsrc": zsrc,
        "massdef": massdef,
        "halo_profile_model": halo_profile_model,
        "zsrc_min": zsrc_min,
        "zsrc_max": zsrc_max,
        "shapenoise": shapenoise,
        "mean_e_err": mean_e_err,
        "photoz_sigma_unscaled": photoz_sigma_unscaled,
        "pz_bins": pz_bins,
        "pz_quantiles_conf": pz_quantiles_conf,
        "field_size": field_size,
        "pzpdf_type": pzpdf_type,
        "coordinate_system": coordinate_system,
    }

    if ngals is None and ngal_density is None:
        err = (
            "Either the number of galaxies 'ngals' or the galaxy density"
            " 'ngal_density' keyword must be specified"
        )
        raise ValueError(err)

    if ngals is not None and ngal_density is not None:
        err = "The 'ngals' and 'ngal_density' keywords cannot both be set. Please use only one."
        raise ValueError(err)

    if ngal_density is not None:
        # Compute the number of galaxies to be drawn
        ngals = _compute_ngals(
            ngal_density,
            field_size,
            cosmo,
            cluster_z,
            zsrc,
            zsrc_min=zsrc_min,
            zsrc_max=zsrc_max,
        )

    galaxy_catalog = _generate_galaxy_catalog(ngals=ngals, **params)
    # Check for bad galaxies and replace them
    nbad, badids = _find_aphysical_galaxies(galaxy_catalog, zsrc_min)
    ntry = 0
    # Prep bins for replacement
    if photoz_sigma_unscaled is not None and pzpdf_type == "shared_bins" and nbad > 0:
        params["pz_bins"] = galaxy_catalog.pzpdf_info["zbins"]
    while (nbad > 0) and (ntry < nretry):
        replacements = _generate_galaxy_catalog(ngals=nbad, **params)
        # galaxy_catalog[badids] = replacements
        for badid, replacement in zip(badids, replacements):
            for col in galaxy_catalog.colnames:
                galaxy_catalog[col][badid] = replacement[col]
        nbad, badids = _find_aphysical_galaxies(galaxy_catalog, zsrc_min)
        ntry += 1

    # Final check to see if there are bad galaxies left
    if nbad > 1:
        warnings.warn(f"Not able to remove {nbad} aphysical objects after {nretry} iterations")

    # Now that the catalog is final, add an id column
    galaxy_catalog["id"] = np.arange(ngals)
    return galaxy_catalog


def _compute_ngals(ngal_density, field_size, cosmo, cluster_z, zsrc, zsrc_min=None, zsrc_max=None):
    """
    A private function that computes the number of galaxies to draw given the user-defined
    field size, galaxy density, cosmology, cluster redshift, galaxy redshift distribution
    and requested redshift range.

    Parameters
    ----------
    ngal_density : float, optional
        The number density of galaxies (in galaxies per square arcminute, from z=0 to z=infty).
        The number of galaxies to be drawn will then depend on the redshift distribution and
        user-defined redshift range.  If specified, the ngals argument will be ignored.
    field_size : float, optional
        The size of the field (field_size x field_size) to be simulated.
        Proper distance in Mpc  at the cluster redshift.
    cosmo: clmm.Cosmology, optional
        Cosmology object.
    cluster_z : float
        Cluster redshift
    zsrc : float or str
        Choose the source galaxy distribution to be fixed or drawn from a predefined distribution.

        * `float` : All sources galaxies at this fixed redshift;
        * `str` : Draws individual source gal redshifts from predefined distribution. Options are:

            * `chang13` - Chang et al. 2013 (arXiv:1305.0793);
            * `desc_srd` - LSST/DESC Science Requirement Document (arxiv:1809.01669);

    zsrc_min : float, optional
        The minimum true redshift of the sources. If photoz errors are included, the observed
        redshift may be smaller than zsrc_min.
    zsrc_max : float, optional
        The maximum true redshift of the sources, apllied when galaxy redshifts are drawn from a
        redshift distribution. If photoz errors are included, the observed redshift may be larger
        than zsrc_max.

    Returns
    -------
    ngals : int
        Number of galaxies to be generated.
    """
    field_size_arcmin = convert_units(field_size, "Mpc", "arcmin", redshift=cluster_z, cosmo=cosmo)
    ngals = ngal_density * field_size_arcmin * field_size_arcmin

    if isinstance(zsrc, float):
        ngals = int(ngals)
    elif zsrc in ("chang13", "desc_srd"):
        z_distrib_func = (
            redshift_distributions.chang2013
            if zsrc == "chang13"
            else redshift_distributions.desc_srd
        )
        # Compute the normalisation for the redshift distribution function (z=[0, inf))
        # z_distrib_func(0, is_cdf=True)=0
        norm = z_distrib_func(np.inf, is_cdf=True)
        # Probability to find the galaxy in the requested redshift range
        prob = (
            z_distrib_func(zsrc_max, is_cdf=True) - z_distrib_func(zsrc_min, is_cdf=True)
        ) / norm
        ngals = int(ngals * prob)
    else:
        raise ValueError(f"zsrc (={zsrc}) must be float, 'chang13' or 'desc_srd'")
    return ngals


def _generate_galaxy_catalog(
    cluster_m,
    cluster_z,
    cluster_c,
    cosmo,
    ngals,
    zsrc,
    cluster_ra=None,
    cluster_dec=None,
    delta_so=None,
    massdef=None,
    halo_profile_model=None,
    zsrc_min=None,
    zsrc_max=None,
    shapenoise=None,
    mean_e_err=None,
    photoz_sigma_unscaled=None,
    pz_bins=101,
    pz_quantiles_conf=(5, 31),
    pzpdf_type="shared_bins",
    coordinate_system="euclidean",
    field_size=None,
):
    """A private function that skips the sanity checks on derived properties. This
    function should only be used when called directly from `generate_galaxy_catalog`.
    For a detailed description of each of the parameters, see the documentation of
    `generate_galaxy_catalog`.
    """
    # Too many local variables (22/15)
    # pylint: disable=R0914

    # Create empty GCData object with the given coordinate system
    galaxy_catalog = GCData(meta={"coordinate_system": coordinate_system})

    # Set the source galaxy redshifts
    _draw_source_redshifts(galaxy_catalog, zsrc, zsrc_min, zsrc_max, ngals)

    # Add photo-z errors and pdfs to source galaxy redshifts
    if photoz_sigma_unscaled is not None:
        galaxy_catalog.pzpdf_info["type"] = pzpdf_type
        _compute_photoz_pdfs(
            galaxy_catalog,
            photoz_sigma_unscaled,
            pz_bins=pz_bins,
            pz_quantiles_conf=pz_quantiles_conf,
        )

    # Draw galaxy positions
    _draw_galaxy_positions(
        galaxy_catalog, ngals, cluster_ra, cluster_dec, cluster_z, cosmo, field_size
    )

    # Compute the shear on each source galaxy
    gamt = compute_tangential_shear(
        galaxy_catalog["r_mpc"],
        mdelta=cluster_m,
        cdelta=cluster_c,
        z_cluster=cluster_z,
        z_src=galaxy_catalog["ztrue"],
        cosmo=cosmo,
        delta_mdef=delta_so,
        halo_profile_model=halo_profile_model,
        massdef=massdef,
        z_src_info="discrete",
    )

    gamx = np.zeros(ngals)

    kappa = compute_convergence(
        galaxy_catalog["r_mpc"],
        mdelta=cluster_m,
        cdelta=cluster_c,
        z_cluster=cluster_z,
        z_src=galaxy_catalog["ztrue"],
        cosmo=cosmo,
        delta_mdef=delta_so,
        halo_profile_model=halo_profile_model,
        massdef=massdef,
        z_src_info="discrete",
    )

    c_cl = SkyCoord(cluster_ra * u.deg, cluster_dec * u.deg, frame="icrs")
    c_gal = SkyCoord(galaxy_catalog["ra"] * u.deg, galaxy_catalog["dec"] * u.deg, frame="icrs")

    # position angle of drawn galaxies w.r.t cluster center
    _, posangle = c_cl.separation(c_gal).rad, c_cl.position_angle(c_gal).rad
    posangle += 0.5 * np.pi  # for right convention

    # corresponding shear1,2 components
    gam1 = -gamt * np.cos(2 * posangle) + gamx * np.sin(2 * posangle)
    gam2 = -gamt * np.sin(2 * posangle) - gamx * np.cos(2 * posangle)

    # instrinsic ellipticities
    e1_intrinsic = 0
    e2_intrinsic = 0

    # Add shape noise to source galaxy shears
    if shapenoise is not None:
        e1_intrinsic = shapenoise * np.random.standard_normal(ngals)
        e2_intrinsic = shapenoise * np.random.standard_normal(ngals)

    # Compute ellipticities
    galaxy_catalog["e1"], galaxy_catalog["e2"] = compute_lensed_ellipticity(
        e1_intrinsic, e2_intrinsic, gam1, gam2, kappa
    )

    cols = ["ra", "dec", "e1", "e2"]
    # if adding uncertainties
    if mean_e_err is not None:
        galaxy_catalog["e_err"] = mean_e_err * np.random.uniform(0.9, 1.1, ngals)
        galaxy_catalog["e1"] = np.random.normal(galaxy_catalog["e1"], galaxy_catalog["e_err"])
        galaxy_catalog["e2"] = np.random.normal(galaxy_catalog["e2"], galaxy_catalog["e_err"])
        cols += ["e_err"]
    cols += ["z", "ztrue"]
    if all(c is not None for c in (photoz_sigma_unscaled, pzpdf_type)):
        if galaxy_catalog.pzpdf_info["type"] == "individual_bins":
            cols += ["pzbins"]
        if galaxy_catalog.pzpdf_info["type"] == "quantiles":
            cols += ["pzquantiles"]
        else:
            cols += ["pzpdf"]

    if coordinate_system == "celestial":
        galaxy_catalog["e2"] *= -1  # flip e2 to match the celestial coordinate system

    return galaxy_catalog[cols]


def _draw_source_redshifts(galaxy_catalog, zsrc, zsrc_min, zsrc_max, ngals):
    """Add source galaxy redshifts, either set to a fixed value or draw from a predefined
    distribution to the source catalog.

    Uses a sampling technique found in Numerical Recipes in C, Chap 7.2: Transformation Method.
    Pulling out random values from a given probability distribution.

    Parameters
    ----------
    galaxy_catalog : clmm.GCData
        Source galaxy catalog
    zsrc : float or str
        Choose the source galaxy distribution to be fixed or drawn from a predefined distribution.
        float : All sources galaxies at this fixed redshift.
        str : Draws individual source gal redshifts from predefined distribution. Options are:

            * `chang13` - Chang et al. 2013 (arXiv:1305.0793);
            * `desc_srd` - LSST/DESC Science Requirement Document (arxiv:1809.01669);

    zsrc_min : float
        The minimum source redshift allowed.
    zsrc_max : float, optional
        If source redshifts are drawn, the maximum source redshift
    ngals : float
        Number of galaxies to generate

    Notes
    -----
    Much of this code in this function was adapted from the Dallas group

    Columns of true and 'measured' photometric redshifts are the same here.
    Redshift photometric errors are then added using _compute_photoz_pdfs.
    """
    # Set zsrc to constant value
    if isinstance(zsrc, float):
        zsrc_list = np.ones(ngals) * zsrc

    # Draw zsrc from Chang et al. 2013
    elif zsrc == "chang13":
        zsrc_list = _draw_random_points_from_distribution(
            zsrc_min, zsrc_max, ngals, redshift_distributions.chang2013
        )

    # Draw zsrc from the distribution used in the DESC SRD (arxiv:1809.01669)
    elif zsrc == "desc_srd":
        zsrc_list = _draw_random_points_from_distribution(
            zsrc_min, zsrc_max, ngals, redshift_distributions.desc_srd
        )

    # Draw zsrc from a uniform distribution between zmin and zmax
    elif zsrc == "uniform":
        zsrc_list = np.random.uniform(zsrc_min, zsrc_max, ngals)

    # Invalid entry
    else:
        raise ValueError(f"zsrc must be a float, chang13 or desc_srd. You set: {zsrc}")

    galaxy_catalog["ztrue"] = zsrc_list
    galaxy_catalog["z"] = zsrc_list


def _compute_photoz_pdfs(
    galaxy_catalog, photoz_sigma_unscaled, pz_bins=101, pz_quantiles_conf=(5, 31)
):
    """Private function to add photo-z errors and PDFs to the mock catalog.

    Parameters
    ----------
    galaxy_catalog : clmm.GCData
        Input galaxy catalog to which photoz PDF will be added
    photoz_sigma_unscaled : float
        Width of the Gaussian PDF, without the (1+z) factor
    pz_bins: int, sequence of scalars or str
        Photo-z pdf bins in the given range. If int, the limits are set automatically.
        If is array, must be the bin edges.
    """
    galaxy_catalog["pzsigma"] = photoz_sigma_unscaled * (1.0 + galaxy_catalog["ztrue"])
    galaxy_catalog["z"] = galaxy_catalog["ztrue"] + galaxy_catalog[
        "pzsigma"
    ] * np.random.standard_normal(len(galaxy_catalog))

    if galaxy_catalog.pzpdf_info["type"] is None:
        return

    zmin = galaxy_catalog["z"] - 10.0 * galaxy_catalog["pzsigma"]
    zmax = galaxy_catalog["z"] + 10.0 * galaxy_catalog["pzsigma"]
    zmin[zmin < 0] = 0.0

    if galaxy_catalog.pzpdf_info["type"] == "shared_bins":
        if isinstance(pz_bins, int):
            galaxy_catalog.pzpdf_info["zbins"] = np.linspace(zmin.min(), zmax.max(), pz_bins)
        else:
            galaxy_catalog.pzpdf_info["zbins"] = np.array(pz_bins)
        galaxy_catalog["pzpdf"] = gaussian(
            galaxy_catalog.pzpdf_info["zbins"],
            galaxy_catalog["z"][:, None],
            galaxy_catalog["pzsigma"][:, None],
        )
    elif galaxy_catalog.pzpdf_info["type"] == "individual_bins":
        if isinstance(pz_bins, int):
            galaxy_catalog["pzbins"] = np.linspace(zmin, zmax, pz_bins).T
        else:
            galaxy_catalog["pzbins"] = [
                pz_bins[max(np.digitize(z1, pz_bins) - 1, 0) : np.digitize(z2, pz_bins) + 1]
                for z1, z2 in zip(zmin, zmax)
            ]
        galaxy_catalog["pzpdf"] = [
            gaussian(row["pzbins"], row["z"], row["pzsigma"]) for row in galaxy_catalog
        ]
    elif galaxy_catalog.pzpdf_info["type"] == "quantiles":
        sigma_steps = np.linspace(-pz_quantiles_conf[0], pz_quantiles_conf[0], pz_quantiles_conf[1])
        galaxy_catalog.pzpdf_info["quantiles"] = 0.5 * erfc(-sigma_steps / np.sqrt(2))
        galaxy_catalog["pzquantiles"] = (
            galaxy_catalog["z"][:, None] + sigma_steps * galaxy_catalog["pzsigma"][:, None]
        )
    else:
        raise ValueError(
            "Value of pzpdf_info['type'] " f"(={galaxy_catalog.pzpdf_info['type']}) " "not valid."
        )


def _draw_galaxy_positions(
    galaxy_catalog, ngals, cluster_ra, cluster_dec, cluster_z, cosmo, field_size
):
    """Draw positions of source galaxies around lens and add them to the source catalog.

    We draw physical x and y positions from uniform distribution with -4 and 4 Mpc of the
    lensing cluster center. We then convert these to RA and DEC using the supplied cosmology

    Parameters
    ----------
    galaxy_catalog : clmm.GCData
        Source galaxy catalog
    ngals : float
        The number of source galaxies to draw
    cluster_ra : float
        The cluster right ascension in degrees
    cluster_dec : float
        The cluster declination in degrees
    cluster_z : float
        The cluster redshift
    cosmo : dict
        Dictionary of cosmological parameters. Must contain at least, Omega_c, Omega_b,
        and H0
    field_size : float
        The size of the field (field_size x field_size) to be simulated around the cluster center.
        Proper distance in Mpc at the cluster redshift.
    """
    lens_distance = cosmo.eval_da(cluster_z)  # Mpc

    # Draw galaxy around (ra,dec)=(0.0)
    galaxy_catalog["x_mpc"] = np.random.uniform(-(field_size / 2.0), field_size / 2.0, size=ngals)
    galaxy_catalog["y_mpc"] = np.random.uniform(-(field_size / 2.0), field_size / 2.0, size=ngals)
    galaxy_catalog["r_mpc"] = np.sqrt(galaxy_catalog["x_mpc"] ** 2 + galaxy_catalog["y_mpc"] ** 2)

    # ra and dec for a cluster in (0,0)
    ra = -np.rad2deg(galaxy_catalog["x_mpc"] / lens_distance)
    dec = np.rad2deg(galaxy_catalog["y_mpc"] / lens_distance)
    coord_gals = SkyCoord(ra * u.deg, dec * u.deg, frame="icrs")
    coord_cl = SkyCoord(0.0 * u.deg, 0.0 * u.deg, frame="icrs")

    # position angle of drawn galaxies w.r.t cluster center in original position coord_cl = (0,0)
    position_angle = coord_cl.position_angle(coord_gals).to(u.deg)

    # separation of drawn galaxies w.r.t cluster center in original position coord_cl = (0,0)
    sep = coord_cl.separation(coord_gals)

    # cluster actual position
    c2_cl = SkyCoord(cluster_ra * u.deg, cluster_dec * u.deg, frame="icrs")

    # new galaxy (ra,dec) w.r.t the new cluster position c2_cl
    new_coord = c2_cl.directional_offset_by(position_angle, sep)

    galaxy_catalog["ra"] = new_coord.ra.degree
    galaxy_catalog["dec"] = new_coord.dec.degree


def _find_aphysical_galaxies(galaxy_catalog, zsrc_min):
    r"""Finds the galaxies that have aphysical derived values due to large systematic choices.

    Currently checks the following conditions
    e1 \in [-1, 1]
    e2 \in [-1, 1]
    z  < zsrc_min
    This was converted to a seperate function to allow for ease of extension without needing
    to change the same code in multiple locations.

    Parameters
    ----------
    galaxy_catalog : clmm.GCData
        Galaxy source catalog
    zsrc_min : float
        Minimum galaxy redshift allowed

    Returns
    -------
    nbad : int
        The number of aphysical galaxies in galaxy_catalog
    badgals : array_like
        A list of the indicies in galaxy_catalog that need to be redrawn
    """
    etot = np.sqrt(
        galaxy_catalog["e1"] * galaxy_catalog["e1"] + galaxy_catalog["e2"] * galaxy_catalog["e2"]
    )

    #     badgals = np.where((np.abs(galaxy_catalog['e1']) > 1.0) |
    #                        (np.abs(galaxy_catalog['e2']) > 1.0) |
    #                        (galaxy_catalog['ztrue'] < zsrc_min)
    #                       )[0]

    badgals = np.where((etot > 1.0) | (galaxy_catalog["ztrue"] < zsrc_min))[0]
    nbad = len(badgals)
    return nbad, badgals
